{% if states %}
<div class="progress">
  <div class="progress-bar bg-success" role="progressbar" style="width: {{states.COMPLETED}}%" aria-valuenow="30" aria-valuemin="0" aria-valuemax="100" title="Completed">{{states.COMPLETED}}%</div>
  <div class="progress-bar bg-warning" role="progressbar" style="width: {{states.PENDING}}%" aria-valuenow="15" aria-valuemin="0" aria-valuemax="100" title="Pending">{{states.PENDING}}%</div>
  <div class="progress-bar bg-danger" role="progressbar" style="width: {{states.FAILED}}%" aria-valuenow="20" aria-valuemin="0" aria-valuemax="100" title="Failed">{{states.FAILED}}%</div>
</div>
<p>Request Success rate</p>
{% else %}
<p>No Observation Requests have been made</p>
{% endif %}
