# Generated by Django 2.2.2 on 2019-09-25 15:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_dataproducts', '0006_auto_20190912_2013'),
        ('tom_education', '0003_asyncprocess_process_type'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='timelapseprocess',
            name='asyncprocess_ptr',
        ),
        migrations.RemoveField(
            model_name='timelapseprocess',
            name='timelapse_product',
        ),
        migrations.CreateModel(
            name='TimelapsePipeline',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('tom_education.pipelineprocess',),
        ),
        migrations.AlterField(
            model_name='asyncprocess',
            name='identifier',
            field=models.CharField(max_length=100, unique=True),
        ),
        migrations.DeleteModel(
            name='TimelapseDataProduct',
        ),
        migrations.DeleteModel(
            name='TimelapseProcess',
        ),
    ]
