{# Need to override this template to set <button> type to "button" to prevent it submitting the timelapse creation form #}
<button type="button" class="btn btn-primary" onclick="openJS9Window('{{ url }}')" title="Open in JS9">View</button>
